﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Analysis.SolutionCaching.RealVectorEncoding {
  [Item("RealVectorCrossoverPointcut", "Wraps a crossover and applies after-crossover actions.")]
  [StorableClass]
  public class RealVectorCrossoverPointcut : CrossoverPointcut<IRealVectorAdvice, IRealVectorCrossover>, IRealVectorCrossover, IRealVectorPointcut {
    public ILookupParameter<ItemArray<RealVector>> ParentsParameter {
      get { return (ScopeTreeLookupParameter<RealVector>)Parameters["Parents"]; }
    }
    public ILookupParameter<RealVector> ChildParameter {
      get { return (ILookupParameter<RealVector>)Parameters["Child"]; }
    }
    public IValueLookupParameter<DoubleMatrix> BoundsParameter {
      get { return (IValueLookupParameter<DoubleMatrix>)Parameters["Bounds"]; }
    }

    [StorableConstructor]
    protected RealVectorCrossoverPointcut(bool deserializing) : base(deserializing) { }
    protected RealVectorCrossoverPointcut(RealVectorCrossoverPointcut original, Cloner cloner) : base(original, cloner) { }

    public RealVectorCrossoverPointcut()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<RealVector>("Parents", "The parent vectors which should be crossed."));
      ParentsParameter.ActualName = "RealVector";
      Parameters.Add(new LookupParameter<RealVector>("Child", "The child vector resulting from the crossover."));
      ChildParameter.ActualName = "RealVector";
      Parameters.Add(new ValueLookupParameter<DoubleMatrix>("Bounds", "The lower and upper bounds of the real vector."));

      //Would be nicer if it comes from operator collection, but MultiRealVector Crossover does the same thing.
      //Also: MultiRealVectorCrossover cannot be used here as it would generate a cyclic recursion 
      //MultiRealVectorCrossover discovers RealVectorCrossoverPointcut which discovers Multi....
      foreach (Type t in ApplicationManager.Manager.GetTypes(typeof(IRealVectorCrossover))) {
        if ((!typeof(RealVectorCrossoverPointcut).IsAssignableFrom(t)) &&
          (!typeof(MultiOperator<IRealVectorCrossover>).IsAssignableFrom(t))) {
          CrossoversParameter.ValidValues.Add((IRealVectorCrossover)Activator.CreateInstance(t));
        }
      }

      ParameterizeCrossovers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RealVectorCrossoverPointcut(this, cloner);
    }

    private void ParameterizeCrossovers() {
      foreach (IRealVectorCrossover crossover in CrossoversParameter.ValidValues.OfType<IRealVectorCrossover>()) {
        crossover.ChildParameter.ActualName = ChildParameter.Name;
        crossover.ParentsParameter.ActualName = ParentsParameter.Name;
        crossover.BoundsParameter.ActualName = BoundsParameter.Name;
      }
      foreach (IStochasticOperator crossover in CrossoversParameter.ValidValues.OfType<IStochasticOperator>()) {
        crossover.RandomParameter.ActualName = RandomParameter.Name;
      }
    }
  }
}
